;;; internals.scm --- spmr implementation

;;; Copyright (C) 2002 by Walter C. Pelissero

;;; Author: Walter C. Pelissero <walter@pelissero.org>
;;; $Id: internals.scm,v 2.6 2003/10/01 17:09:14 wcp Exp $

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2, or
;;; (at your option) any later version.
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;; You should have received a copy of the GNU General Public License
;;; along with this program; see the file COPYING.  If not, write to
;;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;;; Boston, MA 02111-1307, USA.

;;; Commentary:


(define *message* '())		; the complete message
(define *headers* '())		; just the headers
(define *body* '())			; just the body
(define *xloop-id* "nobody@null.dev")
(define *from-in-replies* (make-fluid "MAILER-DELIVERY-AGENT@pelissero.org (spmr.scm)"))
(define *reply-to-in-replies* (make-fluid "MAILER-DELIVERY-AGENT@dont.answer (spmr.scm)"))

;; we must be able to override the uid via environment variables
;; or in the configuration file
(define-lazy user
  (or (getenv "USER") (getenv "LOGNAME") (getenv "USERNAME") (user-login-name)))
(define-lazy mailbox
  (expand-file-name (or (getenv "MAIL")
			(string-append "/var/mail/" (user)))))
(define-lazy user-mail-directory (expand-file-name "~/Mail"))


(define (absolute-filename name)
  (expand-file-name name (user-mail-directory)))

(define (write-list-to-port port list)
  (for-each (lambda (line)
	      (cond ((pair? line)
		     (newline port)
		     (write-list-to-port port line)
		     (newline port))
		    (else
		     (display line port)
		     (newline port))))
	    list))

(define (header->symbol str)
  (string->symbol (string-downcase str)))

(define (symbol->header sym)
  (string-capitalize (symbol->string sym)))

(define (alist->headers l)
  (if (null? l)
      '()
      (let ((h (car l)))
	(if (cdr h)
	    (cons (string-append (symbol->header (car h)) ": " (cdr h))
		  (alist->headers (cdr l)))
	    (alist->headers (cdr l))))))

(define (alookup key alist . default)
  (cond ((assoc key alist) => cdr)
	((null? default) #f)
	(else (car default))))

(define (get-header hdr . default)
  (apply alookup hdr *headers* default))

(define (select-headers alist sel)
  (map (lambda (k)
	 (cons k (alookup k alist "")))
       sel))

(define-syntax header
  (syntax-rules ()
    ((header name)
     (get-header 'name))
    ((header name default)
     (get-header 'name default))))

;; message is a list of lines.  Return two values: the headers and
;; the body.  Headers are represented in an alist.
(define split-headers-from-body
  (let ((splitter (infix-splitter (rx ":" (+ white)) -1))
	;; As some MUAs send empty headers, value is optional
	(cv-header (lambda (hdr . value)
		     (cons (header->symbol (remove-trailing-whitespace hdr))
			   (if (null? value)
			       ""
			       (car value))))))
    (lambda (message)
      (let loop ((msg message)
		 (headers '()))
	(if (null? msg)
	    (values headers '())
	    (let ((line (car msg)))
	      (cond ((string-empty? line)
		     ;; we have found the empty line divides headers
		     ;; from body
		     (values headers (cdr msg)))
		    ;; if the first char is a whitespace this is a
		    ;; continuation line
		    ((char-whitespace? (string-ref line 0))
		     (modify-cdr! (car headers)
				  (lambda (val)
				    (string-append val " " (trim-whitespace line))))
		     (loop (cdr msg) headers))
		    (else
		     (loop (cdr msg) (cons (apply cv-header (splitter line)) headers))))))))))

(define (parse-message port)
  ;; NOTE: The first line of a message is the "From " line.
  (let* ((lines (port->string-list port))
	 (from (car lines)))
    (set! *message* lines)
    (receive (headers body)
	(split-headers-from-body (cdr lines))
      (set! *headers* headers)
      (set! *body* body))))

;; unpack the mime message and returns its parts as an alist of
;; filename and mime type
(define (unpack-mime-message msg)
  (reverse
   (let ((port (run/port (| (begin (write-list *message*))
			    (munpack -t))
			 (- 2))))
     (awk (read-line port) (line) ((parts '()))
	  ((: bos (submatch (+ (~ white))) (+ white)
	      "(" (submatch (+ (~ ")"))) ")") =>
	      (lambda (match)
		(cons (cons (match:substring match 2)
			    (match:substring match 1))
		      parts)))))))

;; we set it inside call-with-mime-parts
(define mime-parts (make-fluid #f))

(define (call-with-mime-parts thunk)
  (if (fluid mime-parts)
      (thunk)
      (within-temp-directory
       (let ((parts (lazy (unpack-mime-message *message*))))
	 (let-fluid mime-parts parts thunk)))))

(define-syntax with-mime-parts
  (syntax-rules ()
    ((with-mime-parts body ...)
     (call-with-mime-parts (thunk body ...)))))

(define (write-list list)
  (write-list-to-port (current-output-port) list))

;; we expect message to be a list of strings
(define (sendmail addresses message)
  (run (| (begin
	    (write-list message))
	  (sendmail ,@addresses))))

(define (merge-headers hl1 hl2)
  (let ((new hl1))
    (for-each (lambda (h)
		(cond ((assoc (car h) new) =>
		       (lambda (e)
			 (set-cdr! e (cdr h))))
		      (else
		       (push h new))))
	      hl2)
    new))

(define (make-answer-headers additional-headers)
  (let ((headers `((subject . ,(header subject ""))
		   (to . ,(reply-address))
		   (from . ,(fluid *from-in-replies*))
		   (reply-to . ,(fluid *reply-to-in-replies*))
		   (x-loop . ,*xloop-id*)))
	(id (header message-id)))
    (alist->headers
     (merge-headers
      (merge-headers (if id
			 `((in-reply-to . ,id) (references . ,id))
			 '())
		     headers)
      additional-headers))))

(define (make-mime-message subject body attachment)
  (let ((desc-file #f)
	(out-file #f))
    (dynamic-wind
     (thunk
      (set! desc-file (create-temp-file))
      (set! out-file (create-temp-file))
      (call-with-output-file desc-file
	(lambda (port)
	  (write-list-to-port port body))))
     (thunk
      (run (mpack -s ,subject -o ,out-file -d ,desc-file
		  ,(expand-file-name attachment)))
      (call-with-input-file out-file port->string-list))
     (thunk
      (delete-file desc-file)
      (delete-file out-file)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (_h~ hdr re)
  (let ((h (get-header hdr)))
    (and h (regexp-search? re h))))

(define-syntax h~
  (syntax-rules ()
    ((h~ hdr re)
     (_h~ 'hdr (rx re)))))

(define *text-re* (rx bos "text/plain"))
(define *html-re* (rx bos "text/html"))

(define (mime-message?)
  (and (header mime-version)
       (header content-type)
       (not (regexp-search? *text-re* (header content-type)))))

(define *on-exit-calls* '())
(define (on-exit proc)
  (push proc *on-exit-calls*))
(define (stop . val)
  (for-each (lambda (proc) (proc))
	    *on-exit-calls*)
  (apply exit val))

(define-lazy rules-file
  (expand-file-name (or (getenv "SPMRRC") "~/.spmr.conf")))

(define (run-rules environment)
  (load (rules-file) (get-package environment)))

(define (map-mime-parts proc)
  (with-mime-parts
   (map (lambda (part)
	  ;; (proc encoding file)
	  (proc (car part) (cdr part)))
	((fluid mime-parts)))))

(define (for-each-mime-part proc)
  (with-mime-parts
   (for-each (lambda (part)
	       ;; (proc encoding file)
	       (proc (car part) (cdr part)))
	     ((fluid mime-parts)))))

(define-lazy printable-message
  (if (mime-message?)
      (reverse
       (with-mime-parts
	(let loop ((parts ((fluid mime-parts)))
		   (lines '()))
	  (if (null? parts)
	      lines
	      (let ((encoding (caar parts))
		    (file (cdar parts)))
		(cond ((regexp-search? *text-re* encoding)
		       (loop (cdr parts)
			     (append (reverse (file->strings file)) lines)))
		      ((regexp-search? *html-re* encoding)
		       (let ((content (port->string-list (run/port (lynx -dump -force_html ,file)))))
			 (loop (cdr parts)
			       (append (list "" (string-append "[ end of MIME attachment \"" file "\" ]") "")
				       (reverse content)
				       (list "" (string-append "[ start of MIME attachment \"" file
							       "\" of type " encoding " ]") "")
				       lines))))
		      (else
		       (loop (cdr parts)
			     (cons (string-append "[[ omitted MIME attachment \"" file
						  "\" of type " encoding " ]]")
				   lines)))))))))
      *body*))

(define-lazy quoted-message
  (map (curry string-append " > ")
       (printable-message)))

(define-lazy reply-address
  (or (header reply-to) (header from)))

(define (print-message)
  (let ((print-line (thunk
		     (display (make-string 75 #\-))
		     (newline)))
	(print-headers (lambda (sel)
			 (write-list (alist->headers (select-headers *headers* sel)))))
	(print-body (thunk
		     (write-list (printable-message)))))
    (run (| (begin
	      (print-headers '(from subject date))
	      (newline)
	      (print-body))
	    (lpr)))))

(define (blacklist list)
  (with-exclapp (port (absolute-filename list))
    (display (header from) port)
    (newline port)))

(define (blacklisted? list)
  (let ((sender (header from)))
    (call-with-input-file (absolute-filename list)
      (lambda (port)
	(let loop ((line (read-line port)))
	  (if (eof-object? line)
	      #f
	      (or (string=? line sender)
		  (loop (read-line port)))))))))

(define (looping-message?)
  (h~ x-loop ,*xloop-id*))

(define (answer-mime body attachment)
  (if (not (looping-message?))
      (sendmail (list (reply-address))
		(append (make-answer-headers '((subject . #f)))
			(make-mime-message
			 (header subject)
			 (append (list body "" (string-append (header from) " wrote:"))
				 (quoted-message))
			 attachment)))))

(define (answer body)
  (if (not (looping-message?))
      (sendmail (list (reply-address))
		(append (make-answer-headers '())
			(list "" body "" (string-append (header from) " wrote:"))
			(quoted-message)))))

(define (answer-once db body)
  (when (not (blacklisted? db))
    (answer body)
    (blacklist db)))

(define (forward . addresses)
  (sendmail addresses *message*))

(define (forward-to-program prog . args)
  (run (| (begin
	    (write-list *message*))
	  (,prog ,@args))))

(define (save . file)
  (if (null? file)
      (save (mailbox))
      (with-exclapp (port (absolute-filename (car file)))
        (write-list-to-port port *message*)
	(newline port))))

;;;
;;; Leave the following lines at end of file.
;;;
;;; Local Variables:
;;; mode:scheme
;;; scheme-program-name:"scsh"
;;; End:
