/*  spmr.c --- spmr trampoline

 Copyright (C) 2003 by Walter C. Pelissero

 Author: Walter C. Pelissero <walter@pelissero.de>
 $Id: spmr.c,v 1.1 2003/11/20 15:46:52 wcp Exp $

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License as
 published by the Free Software Foundation; either version 2, or
 (at your option) any later version.
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this program; see the file COPYING.  If not, write to
 the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.

 Commentary:

This tiny C program ensures that only one instance of the SPMR image
is running a time per user. Normally a new instance of the SCSH
virtual machine is started for each incoming mail message and this can
be rather heavy on small/medium systems.  SCSH may not be fast enough
to bootstrap and process each message.  If SPMR doesn't keep up with
the incoming mails the system is going to be clogged with SCSH process
bringing the whole system do a grinding halt.  This program avoids
this situation because its very small memory requirements makes it
easy for any system to have several instances around without any
performance penalty.

The trampoline will run the real SPMR only if an advisory lock can be
applied on the user's SPMR config file.  Otherwise it will sit waiting
for SPMR to finish. */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <pwd.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>

/* #define DEBUG */
#define LOCK_FILE_TEMPLATE "%s/.spmr.conf"
#define SPMR_IMAGE "spmr.image"

static const char RCSid[] = "$Id: spmr.c,v 1.1 2003/11/20 15:46:52 wcp Exp $";

static const char *myname;
static const char *login_name;
static const char *home_dir;
static const char *lock_file_name;


static struct passwd *
pw_data ()
{
  static struct passwd *pw;
  static uid_t uid;

  if (pw)
    return pw;
  uid = getuid();
  pw = getpwuid(uid);
  if (!pw)
    {
      fprintf(stderr, "%s: uid %d not present in passwd file?!\n", myname, uid);
      exit(1);
    }
  endpwent();
  return pw;
}


static void
load_user_data ()
{
  if (login_name)
    return;
  if (!(login_name = getenv("USER")))
    login_name = pw_data()->pw_name;
  if (!(home_dir = getenv("HOME")))
    home_dir = pw_data()->pw_dir;
  {
    char buf[1024];

    sprintf(buf, LOCK_FILE_TEMPLATE, home_dir);
    lock_file_name = strdup(buf);
  }
#ifdef DEBUG
  fprintf(stderr, "login_name = %s, home_dir = %s, lock_file_name = %s\n",
	  login_name, home_dir, lock_file_name);
#endif
}

static int
get_lock ()
{
  int fd;

#ifdef __FreeBSD__
  if ((fd = open(lock_file_name, O_RDONLY | O_EXLOCK)) == -1)
    {
      fprintf(stderr, "%s: cannot open %s", myname, lock_file_name);
      exit(1);
    }
#endif
#ifdef __Linux__
  if ((fd = open(lock_file_name, O_RDONLY)) == -1)
    {
      fprintf(stderr, "%s: cannot open %s", myname, lock_file_name);
      exit(1);
    }
  flock(fd, LOCK_EX);
#endif
#ifdef DEBUG
  fprintf(stderr, "lock obtained: fd = %d\n", fd);
#endif
  return fd;
}

static void
run_spmr_image ()
{
  pid_t pid;

  if (!getenv("HOME"))
    setenv("HOME", home_dir, 1);
  if (!getenv("USER"))
    setenv("USER", login_name, 1);
  if ((pid = fork()) == 0)
    {
      char image[1024];

      sprintf(image, "%s.image", myname);
      if (myname[0] == '/')
	execl(image, image, 0);
      else
	execlp(image, image, 0);
      fprintf(stderr, "%s: cannot execute %s", myname, image);
      exit(1);
    }
  else if (pid > 0)
    {
      int status;
      wait(&status);
    }
  else
    {
      fprintf(stderr, "%s: cannot fork\n", myname);
      exit(1);
    }
}

static void
release_lock (int fd)
{
  close(fd);
#ifdef DEBUG
  fprintf(stderr, "lock released: fd = %d\n", fd);
#endif
}

int
main (int argc, const char *argv[])
{
  int lock;

  myname = argv[0];
  load_user_data();
  lock = get_lock();
  run_spmr_image();
  release_lock(lock);
  return 0;
}
