;;; utils.scm --- various functions

;;; Copyright (C) 2002 by Walter C. Pelissero

;;; Author: Walter C. Pelissero <walter@pelissero.org>
;;; $Id: utils.scm,v 1.6 2003/03/27 20:33:36 wcp Exp $

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2, or
;;; (at your option) any later version.
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;; You should have received a copy of the GNU General Public License
;;; along with this program; see the file COPYING.  If not, write to
;;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;;; Boston, MA 02111-1307, USA.

;;; Commentary:


(define-syntax lazy
  (syntax-rules ()
    ((lazy body)
     (let ((promise (delay body)))
       (lambda ()
	 (force promise))))))

(define-syntax define-lazy
  (syntax-rules ()
    ((define-lazy name body)
     (define name (lazy body)))))

(define-lazy tmp-dir (or (getenv "TMPDIR") "/var/tmp"))

(define-syntax thunk
  (syntax-rules ()
    ((thunk body ...)
     (lambda ()
       body ...))))

(define (curry proc . args)
  (lambda other-args
    (apply proc (append args other-args))))

(define (modify-cdr! pair function)
  (set-cdr! pair (function (cdr pair))))

(define-syntax when
  (syntax-rules ()
    ((when condition body ...)
     (if condition
	 (begin body ...)))))

(define-syntax push
  (syntax-rules ()
    ((push item list)
     (begin
       (set! list (cons item list))
       list))))

(define call/cc call-with-current-continuation)

(define-syntax w/cc
  (syntax-rules ()
    ((w/cc name body ...)
     (call/cc
      (lambda (name)
	body ...)))))

(define (member-if pred l)
  (let loop ((l l))
    (cond ((null? l)
	   #f)
	  ((pred (car l))
	   l)
	  (else
	   (loop (cdr l))))))

(define (string-empty? str)
  (zero? (string-length str)))

(define (negate pred)
  (lambda args
    (not (apply pred args))))

(define (bior . args)
  (reduce bitwise-ior 0 args))

(define (call-with-open-file file mode proc)
  (let ((port #f))
    (dynamic-wind
     (lambda ()
       (set! port (open-file file mode)))
     (lambda ()
       (proc port))
     (lambda ()
       (close port)))))

(define-syntax with-open-file
  (syntax-rules ()
    ((_ (var name mode) body ...)
     (call-with-open-file name mode
			  (lambda (var)
			    body
			    ...)))))

(define-syntax with-exclapp
  (syntax-rules ()
    ((_ (var name) body ...)
     (with-open-file (var name (bior open/create open/write open/append open/exclusive-lock))
       body ...))))

(define (remove-leading-whitespace str)
  (list->string
   (or (member-if (negate char-whitespace?) (string->list str))
       '())))

(define (remove-trailing-whitespace str)
  (list->string (reverse (or (member-if (negate char-whitespace?)
					(reverse (string->list str)))
			     '()))))

(define (string-capitalize str)
  (list->string
   (let loop ((prev-was-letter #f)
	      (l (string->list str)))
     (if (null? l)
	 '()
	 (let ((next (car l)))
	   (cons (if prev-was-letter
		     next
		     (char-upcase next))
		 (loop (char-alphabetic? next) (cdr l))))))))

(define (trim-whitespace str)
  (remove-trailing-whitespace (remove-leading-whitespace str)))

(define (make-temp-dir base)
  (temp-file-iterate (lambda (dir) (create-directory dir) dir)
		     (string-append base "/spmr." (number->string (pid)) "~a")))

(define (call-within-temp-directory proc)
  (let ((tmpdir #f))
    (dynamic-wind
     (lambda ()
       (set! tmpdir (make-temp-dir (tmp-dir))))
     (lambda ()
       (with-cwd tmpdir
		 (proc)))
     (lambda ()
       (run (rm -rf ,tmpdir))))))

(define-syntax within-temp-directory
  (syntax-rules ()
    ((within-temp-directory body ...)
     (call-within-temp-directory (lambda () body ...)))))

(define (file->strings name)
  (call-with-input-file name port->string-list))

(define (file->string name)
  (call-with-input-file name port->string))

;;;
;;; Leave the following lines at end of file.
;;;
;;; Local Variables:
;;; mode:scheme
;;; scheme-program-name:"scsh"
;;; End:
